import time
import traceback
import typing
from collections.abc import Callable
from enum import StrEnum

import notte_core
from litellm import AllMessageValues, override
from loguru import logger
from notte_browser.session import NotteSession
from notte_browser.vault import VaultSecretsScreenshotMask
from notte_browser.window import BrowserWindow
from notte_core.actions import (
    BaseAction,
    CompletionAction,
    FallbackObserveAction,
)
from notte_core.browser.observation import Observation
from notte_core.common.config import NotteConfig, RaiseCondition
from notte_core.common.tracer import LlmUsageDictTracer
from notte_core.credentials.base import BaseVault, LocatorAttributes
from notte_core.llms.engine import LLMEngine
from notte_core.utils.webp_replay import ScreenshotReplay, WebpReplay
from notte_sdk.types import AgentCreateRequestDict, AgentRunRequest, AgentRunRequestDict
from patchright.async_api import Locator
from pydantic import computed_field, field_validator

from notte_agent.common.base import BaseAgent
from notte_agent.common.captcha_detector import CaptchaDetector
from notte_agent.common.conversation import Conversation
from notte_agent.common.safe_executor import ExecutionStatus, SafeActionExecutor
from notte_agent.common.trajectory_history import TrajectoryStep
from notte_agent.common.types import AgentResponse
from notte_agent.common.validator import CompletionValidator
from notte_agent.falco.perception import FalcoPerception
from notte_agent.falco.prompt import FalcoPrompt
from notte_agent.falco.trajectory_history import FalcoTrajectoryHistory
from notte_agent.falco.types import StepAgentOutput

# TODO: list
# handle tooling calling methods for different providers (if not supported by litellm)
# Handle control flags
# Done callback
# Setup telemetry
# Setup memory
# Handle custom functions, e.g. `Upload file to element`ç
# Remove base 64 images from current state
# TODO: add fault tolerance LLM parsing
# TODO: only display modal actions when modal is open (same as before)
# TODO: handle prevent default click JS events
# TODO: add some tree structure for menu elements (like we had in notte before. Ex. Menu in Arxiv)


class HistoryType(StrEnum):
    FULL_CONVERSATION = "full_conversation"
    SHORT_OBSERVATIONS = "short_observations"
    SHORT_OBSERVATIONS_WITH_RAW_DATA = "short_observations_with_raw_data"
    SHORT_OBSERVATIONS_WITH_SHORT_DATA = "short_observations_with_short_data"
    COMPRESSED = "compressed"


class FalcoConfig(NotteConfig):
    enable_perception: bool = False
    auto_scrape: bool = False
    history_type: HistoryType = HistoryType.SHORT_OBSERVATIONS_WITH_SHORT_DATA

    @field_validator("enable_perception")
    def check_perception(cls, value: bool) -> bool:
        if value:
            logger.warning("Perception should be disabled for falco. Don't set this argument to `True`.")
        return False

    @field_validator("auto_scrape")
    def check_auto_scrape(cls, value: bool) -> bool:
        if value:
            logger.warning("Auto scrape is not allowed for falco. Don't set this argument to another value.")
        return False


class FalcoResponse(AgentResponse):
    agent_trajectory: list[TrajectoryStep[StepAgentOutput]]  # pyright: ignore [reportIncompatibleVariableOverride]

    @computed_field
    @property
    def steps(self) -> list[StepAgentOutput]:
        return [step.agent_response for step in self.agent_trajectory]

    @override
    def replay(self) -> WebpReplay:
        screenshots: list[bytes] = []
        texts: list[str] = []

        for step in self.agent_trajectory:
            for obs in step.observations():
                if obs.screenshot is not None:
                    screenshots.append(obs.screenshot)
                    texts.append(step.agent_response.state.next_goal)

        if len(screenshots) == 0:
            raise ValueError("No screenshots found in agent trajectory")

        return ScreenshotReplay.from_bytes(screenshots).get(step_text=texts)  # pyright: ignore [reportArgumentType]


class FalcoAgent(BaseAgent):
    def __init__(
        self,
        window: BrowserWindow,
        vault: BaseVault | None = None,
        step_callback: Callable[[str, StepAgentOutput], None] | None = None,
        **data: typing.Unpack[AgentCreateRequestDict],
    ):
        self.config: FalcoConfig = FalcoConfig.from_toml(**data)
        session = NotteSession(window=window, enable_perception=False)
        super().__init__(session=session)
        self.vault: BaseVault | None = vault

        self.tracer: LlmUsageDictTracer = LlmUsageDictTracer()
        self.llm: LLMEngine = LLMEngine(model=self.config.reasoning_model, tracer=self.tracer)

        self.step_callback: Callable[[str, StepAgentOutput], None] | None = step_callback
        # Users should implement their own parser to customize how observations
        # and actions are formatted for their specific LLM and use case

        if self.vault is not None:
            # hide vault leaked credentials within llm inputs
            self.llm.structured_completion = self.vault.patch_structured_completion(0, self.vault.get_replacement_map)(  # pyright: ignore [reportAttributeAccessIssue]
                self.llm.structured_completion
            )

        self.perception: FalcoPerception = FalcoPerception()
        self._validator: CompletionValidator | None = None
        self.captcha_detector: CaptchaDetector = CaptchaDetector(llm=self.llm, perception=self.perception)
        self.prompt: FalcoPrompt = FalcoPrompt(max_actions_per_step=self.config.max_actions_per_step)
        self.conv: Conversation = Conversation(
            convert_tools_to_assistant=True,
            autosize=True,
            model=self.config.reasoning_model,
        )
        self.trajectory: FalcoTrajectoryHistory = FalcoTrajectoryHistory(max_steps=self.config.max_steps)

        async def execute_action(action: BaseAction) -> Observation:
            if self.vault is not None and self.vault.contains_credentials(action):
                locator = await self.session.locate(action)
                if locator is not None:
                    action = self.vault.replace_credentials(
                        action,
                        await FalcoAgent.compute_locator_attributes(locator),
                        self.session.snapshot,
                    )
            _ = await self.session.astep(action)
            return await self.session.aobserve()

        self.step_executor: SafeActionExecutor[BaseAction, Observation] = SafeActionExecutor(func=execute_action)

    @property
    def validator(self) -> CompletionValidator:
        if self._validator is not None:
            return self._validator

        self._validator = CompletionValidator(
            llm=self.llm, perception=self.perception, use_vision=self.config.use_vision
        )
        return self._validator

    @staticmethod
    async def compute_locator_attributes(locator: Locator) -> LocatorAttributes:
        attr_type = await locator.get_attribute("type")
        autocomplete = await locator.get_attribute("autocomplete")
        outer_html = await locator.evaluate("el => el.outerHTML")
        return LocatorAttributes(type=attr_type, autocomplete=autocomplete, outerHTML=outer_html)

    async def reset(self) -> None:
        self.conv.reset()
        self.trajectory.reset()
        self.step_executor.reset()
        await self.session.areset()

    def output(self, answer: str, success: bool) -> AgentResponse:
        return FalcoResponse(
            answer=answer,
            success=success,
            session_trajectory=self.session.trajectory,
            agent_trajectory=self.trajectory.steps,
            messages=self.conv.messages(),
            duration_in_s=time.time() - self.start_time,
            llm_usage=self.tracer.usage,
        )

    async def get_messages(self, task: str) -> list[AllMessageValues]:
        self.conv.reset()
        system_msg, task_msg = self.prompt.system(), self.prompt.task(task)
        if self.vault is not None:
            system_msg += "\n" + self.vault.instructions()
        self.conv.add_system_message(content=system_msg)
        self.conv.add_user_message(content=task_msg)
        # just for logging
        traj_msg = self.trajectory.perceive()
        if self.config.verbose:
            logger.trace(f"🔍 Trajectory history:\n{traj_msg}")
        # add trajectory to the conversation
        match self.config.history_type:
            case HistoryType.COMPRESSED:
                self.conv.add_user_message(content=traj_msg)
            case _:
                if len(self.trajectory.steps) == 0:
                    self.conv.add_user_message(content=self.trajectory.start_rules())
                for step in self.trajectory.steps:
                    # TODO: choose if we want this to be an assistant message or a tool message
                    # self.conv.add_tool_message(step.agent_response, tool_id="step")
                    self.conv.add_assistant_message(step.agent_response.model_dump_json(exclude_none=True))
                    for result in step.results:
                        short_step_msg = self.trajectory.perceive_step_result(result, include_ids=True)
                        self.conv.add_user_message(content=short_step_msg)
                        if not result.success:
                            continue
                        # add observation data to the conversation
                        obs = result.get()
                        match (self.config.history_type, obs.has_data()):
                            case (HistoryType.FULL_CONVERSATION, _):
                                self.conv.add_user_message(
                                    content=self.perception.perceive(obs),
                                    image=(obs.screenshot if self.config.use_vision else None),
                                )
                            case (HistoryType.SHORT_OBSERVATIONS_WITH_RAW_DATA, True):
                                # add data if data was scraped
                                self.conv.add_user_message(content=self.perception.perceive_data(obs, raw=True))

                            case (HistoryType.SHORT_OBSERVATIONS_WITH_SHORT_DATA, True):
                                self.conv.add_user_message(content=self.perception.perceive_data(obs, raw=False))
                            case _:
                                pass

        last_valid_obs = self.trajectory.last_obs()
        if last_valid_obs is not None and self.config.history_type is not HistoryType.FULL_CONVERSATION:
            self.conv.add_user_message(
                content=self.perception.perceive(last_valid_obs),
                image=(last_valid_obs.screenshot if self.config.use_vision else None),
            )

        if len(self.trajectory.steps) > 0:
            self.conv.add_user_message(self.prompt.action_message())

        return self.conv.messages()

    async def step(self, task: str) -> CompletionAction | None:
        """Execute a single step of the agent"""
        messages = await self.get_messages(task)
        response: StepAgentOutput = await self.llm.structured_completion(
            messages, response_format=StepAgentOutput, use_strict_response_format=False
        )

        if self.step_callback is not None:
            self.step_callback(task, response)

        if self.config.verbose:
            logger.trace(f"🔍 LLM response:\n{response}")

        for text, data in response.log_state():
            logger.opt(colors=True).info(text, **data)

        self.trajectory.add_output(response)
        # check for completion
        if response.output is not None:
            return response.output
        # Execute the actions
        for action in response.get_actions():
            result = await self.step_executor.execute(action)

            self.trajectory.add_step(result)
            step_msg = self.trajectory.perceive_step_result(result, include_ids=True)
            logger.info(f"{step_msg}\n\n")
            if not result.success:
                # observe again
                obs = await self.session.aobserve()

                # cast is necessary because we cant have covariance
                # in ExecutionStatus
                ex_status = ExecutionStatus(
                    input=typing.cast(BaseAction, FallbackObserveAction()),
                    output=obs,
                    success=True,
                    message="Observed",
                )
                self.trajectory.add_output(response)
                self.trajectory.add_step(ex_status)

                # stop the loop
                break
            # Successfully executed the action
        return None

    @override
    async def run(self, **kwargs: typing.Unpack[AgentRunRequestDict]) -> AgentResponse:
        request = AgentRunRequest.model_validate(kwargs)
        logger.trace(f"Running task: {request.task}")
        self.start_time: float = time.time()
        try:
            return await self._run(request)

        except Exception as e:
            if self.config.raise_condition is RaiseCondition.NEVER:
                return self.output(f"Failed due to {e}: {traceback.format_exc()}", False)
            raise e

    async def _human_in_the_loop(self) -> None:
        # Check for captcha if human-in-the-loop is enabled
        captcha_result = await self.captcha_detector.detect(self.session.trajectory[-1])
        if captcha_result.has_captcha:
            logger.info(f"⚠️ Captcha detected: {captcha_result.description}")
            logger.info("🔄 Waiting for human intervention...")
            _ = input("Press Enter to continue after solving the captcha...")
            # Observe again after human intervention
            obs = await self.session.aobserve()
            self.trajectory.add_step(
                ExecutionStatus(
                    input=typing.cast(BaseAction, FallbackObserveAction()),
                    output=obs,
                    success=True,
                    message="Observed after human intervention",
                )
            )

    async def _run(self, request: AgentRunRequest) -> AgentResponse:
        """Execute the task with maximum number of steps"""
        # change this to DEV if you want more explicit error messages
        # when you are developing your own agent
        notte_core.set_error_mode("agent")
        if request.url is not None:
            request.task = f"Start on '{request.url}' and {request.task}"

        # hide vault leaked credentials within screenshots
        if self.vault is not None:
            self.session.window.screenshot_mask = VaultSecretsScreenshotMask(vault=self.vault)

        for step in range(self.config.max_steps):
            logger.info(f"💡 Step {step}")
            output: CompletionAction | None = await self.step(task=request.task)
            # Check for captcha if human-in-the-loop is enabled
            if self.config.human_in_the_loop:
                await self._human_in_the_loop()
            if output is None:
                continue
            # validate the output
            if not output.success:
                logger.error(f"🚨 Agent terminated early with failure: {output.answer}")
                return self.output(output.answer, False)
            # Sucessful execution and LLM output is not None
            # Need to validate the output
            logger.info(f"🔥 Validating agent output:\n{output.model_dump_json()}")
            val = await self.validator.validate(
                task=request.task,
                output=output,
                history=self.trajectory,  # pyright: ignore [reportArgumentType]
                response_format=request.response_format,
            )
            if val.is_valid:
                # Successfully validated the output
                logger.info("✅ Task completed successfully")
                return self.output(output.answer, output.success)
            else:
                # TODO handle that differently
                failed_val_msg = f"""Final validation failed: {val.reason}. Continuing...
                CRITICAL: If you think this validation is wrong: argue why the task if finished, or
                perform actions that would prove it is.
                """
                logger.error(failed_val_msg)
                # add the validation result to the trajectory and continue
                self.trajectory.add_step(
                    ExecutionStatus(
                        input=output,
                        output=None,
                        success=False,
                        message=failed_val_msg,
                    )
                )

        error_msg = f"Failed to solve task in {self.config.max_steps} steps"
        logger.info(f"🚨 {error_msg}")
        notte_core.set_error_mode("developer")
        return self.output(error_msg, False)



from typing_extensions import override

from notte_agent.common.trajectory_history import (
    TrajectoryHistory,
    TrajectoryStep,
)
from notte_agent.falco.types import StepAgentOutput


class FalcoTrajectoryHistory(TrajectoryHistory[StepAgentOutput]):
    @override
    def perceive_step(
        self,
        step: TrajectoryStep[StepAgentOutput],
        step_idx: int = 0,
        include_ids: bool = False,
        include_data: bool = True,
    ) -> str:
        action_msg = "\n".join(["  - " + result.input.model_dump_agent_json() for result in step.results])
        status_msg = "\n".join(
            ["  - " + self.perceive_step_result(result, include_ids, include_data) for result in step.results]
        )
        return f"""
# Execution step {step_idx}
* state:
    - page_summary: {step.agent_response.state.page_summary}
    - previous_goal_status: {step.agent_response.state.previous_goal_status}
    - previous_goal_eval: {step.agent_response.state.previous_goal_eval}
    - memory: {step.agent_response.state.memory}
    - next_goal: {step.agent_response.state.next_goal}
* selected actions:
{action_msg}
* execution results:
{status_msg}"""

    @override
    def add_output(self, output: StepAgentOutput) -> None:
        self.steps.append(TrajectoryStep(agent_response=output, results=[]))



from collections.abc import Awaitable
from typing import Callable, Generic, TypeVar, final

from notte_core.common.config import RaiseCondition, config
from notte_core.errors.base import NotteBaseError
from notte_core.errors.provider import RateLimitError
from pydantic import BaseModel
from pydantic_core import ValidationError

S = TypeVar("S")  # Source type
T = TypeVar("T")  # Target type


class ExecutionStatus(BaseModel, Generic[S, T]):
    input: S
    output: T | None
    success: bool
    message: str

    def get(self) -> T:
        if self.output is None or not self.success:
            raise ValueError(f"Execution failed with message: {self.message}")
        return self.output


class StepExecutionFailure(NotteBaseError):
    def __init__(self, message: str):
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )


class MaxConsecutiveFailuresError(NotteBaseError):
    def __init__(self, max_failures: int):
        self.max_failures: int = max_failures
        message = f"Max consecutive failures reached in a single step: {max_failures}."
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )


@final
class SafeActionExecutor(Generic[S, T]):
    def __init__(
        self,
        func: Callable[[S], Awaitable[T]],
        max_consecutive_failures: int = config.max_consecutive_failures,
        raise_on_failure: bool = config.raise_condition is RaiseCondition.IMMEDIATELY,
    ) -> None:
        self.func = func
        self.max_consecutive_failures = max_consecutive_failures
        self.consecutive_failures = 0
        self.raise_on_failure = raise_on_failure

    def reset(self) -> None:
        self.consecutive_failures = 0

    def on_failure(self, input_data: S, error_msg: str, e: Exception) -> ExecutionStatus[S, T]:
        self.consecutive_failures += 1
        if self.consecutive_failures >= self.max_consecutive_failures:
            raise MaxConsecutiveFailuresError(self.max_consecutive_failures) from e
        if self.raise_on_failure:
            raise StepExecutionFailure(error_msg) from e
        return ExecutionStatus(
            input=input_data,
            output=None,
            success=False,
            message=error_msg,
        )

    async def execute(self, input_data: S) -> ExecutionStatus[S, T]:
        try:
            result = await self.func(input_data)
            self.consecutive_failures = 0
            return ExecutionStatus(
                input=input_data,
                success=True,
                output=result,
                message=f"Successfully executed action with input: {input_data}",
            )
        except RateLimitError as e:
            return self.on_failure(input_data, "Rate limit reached. Waiting before retry.", e)
        except NotteBaseError as e:
            # When raise_on_failure is True, we use the dev message to give more details to the user
            msg = e.dev_message if self.raise_on_failure else e.agent_message
            return self.on_failure(input_data, msg, e)
        except ValidationError as e:
            return self.on_failure(
                input_data,
                (
                    "JSON Schema Validation error: The output format is invalid. "
                    f"Please ensure your response follows the expected schema. Details: {str(e)}"
                ),
                e,
            )
        except Exception as e:
            return self.on_failure(input_data, f"An unexpected error occurred: {e}", e)



from typing import final

import chevron
from notte_core.actions import CompletionAction
from notte_core.browser.observation import Observation
from notte_core.llms.engine import LLMEngine
from pydantic import BaseModel, ValidationError

from notte_agent.common.conversation import Conversation
from notte_agent.common.perception import BasePerception
from notte_agent.common.trajectory_history import TrajectoryHistory

system_rules = """
You are a validator of an agent who interacts with a browser.
Validate if the output of last action is what the user wanted and if the task is completed.
If the task is unclear defined, you can let it pass.
But if something is missing or the image does not show what was requested dont let it pass.
Try to understand the page and help the model with suggestions like scroll, do x, ... to get the solution right.

Task to validate: {{task}}.

Return a JSON object with 2 keys: `is_valid` and `reason`:
- `is_valid` is a boolean that indicates if the output is correct.
- `reason` is a string that explains why it is valid or not.

Example:
```json
{{&example}}
```

Your turn:
"""


class CompletionValidation(BaseModel):
    is_valid: bool
    reason: str


@final
class CompletionValidator:
    def __init__(
        self,
        llm: LLMEngine,
        perception: BasePerception,
        use_vision: bool = True,
        include_attributes: bool = True,
        max_steps: int = 3,
    ):
        self.use_vision = use_vision
        self.include_attributes = include_attributes
        self.llm: LLMEngine = llm
        self.conv: Conversation = Conversation()
        self.perception: BasePerception = perception
        self.max_actions: int = max_steps

    @staticmethod
    def example() -> CompletionValidation:
        return CompletionValidation(
            is_valid=False,
            reason="The user wanted to search for 'cat photos', but the agent searched for 'dog photos' instead.",
        )

    def validation_message(
        self, output: CompletionAction, history: TrajectoryHistory[BaseModel], last_obs: Observation
    ) -> str:
        previous_results = [result for step in history.steps for result in step.results][-self.max_actions :]

        return f"""
Last observation:
{self.perception.perceive(last_obs)}


Last action executions:
{"/n".join(TrajectoryHistory.perceive_execution_result(result) for result in previous_results)}

Agent task output:
{output}
"""

    @staticmethod
    def validate_response_format(output: CompletionAction, response_format: type[BaseModel]) -> CompletionValidation:
        """Check that json output fits json schema"""
        try:
            _ = response_format.model_validate_json(output.answer)
        except ValidationError as e:
            return CompletionValidation(is_valid=False, reason=str(e))
        return CompletionValidation(
            is_valid=True, reason="The output returned by the agent is a valid according to the response format."
        )

    async def validate(
        self,
        task: str,
        output: CompletionAction,
        history: TrajectoryHistory[BaseModel],
        response_format: type[BaseModel] | None = None,
    ) -> CompletionValidation:
        """Validate the output of the last action is what the user wanted"""

        # first, validate the output if provided a schema
        if response_format is not None:
            validation = CompletionValidator.validate_response_format(output, response_format)
            if not validation.is_valid:
                return validation

        observations = history.observations()
        if len(observations) == 0:
            return CompletionValidation(is_valid=True, reason="No observations")

        # then, validate that the answer is correct
        last_obs = observations[-1]

        self.conv.reset()
        system_prompt = chevron.render(system_rules, {"task": task, "example": self.example().model_dump_json()})
        self.conv.add_system_message(content=system_prompt)

        validation_message = self.validation_message(output, history, last_obs)

        self.conv.add_user_message(
            content=validation_message,
            image=(last_obs.screenshot if self.use_vision else None),
        )

        answer: CompletionValidation = await self.llm.structured_completion(self.conv.messages(), CompletionValidation)
        return answer



from typing import final

import chevron
from notte_browser.session import TrajectoryStep
from notte_core.llms.engine import LLMEngine
from pydantic import BaseModel

from notte_agent.common.conversation import Conversation
from notte_agent.common.perception import BasePerception

system_rules = """
You are a captcha detector for web pages.
Analyze the provided screenshot and determine if there is a captcha present on the page.
A captcha can be in various forms such as:
- Image-based challenges
- Text-based challenges
- Checkbox-based verification
- Audio-based challenges
- Math problems
- Slider puzzles
- Blocked by network security
- etc.

Return a JSON object with 3 keys: `has_captcha`, `captcha_type`, and `description`:
- `has_captcha` is a boolean indicating if a captcha is present
- `captcha_type` is a string describing the type of captcha (or "none" if no captcha)
- `description` is a string providing details about the captcha's appearance and location

Example:
```json
{{&example}}
```

Your turn:
"""


class CaptchaDetection(BaseModel):
    has_captcha: bool
    captcha_type: str
    description: str


@final
class CaptchaDetector:
    def __init__(
        self,
        llm: LLMEngine,
        perception: BasePerception,
        use_vision: bool = True,
        include_attributes: bool = True,
    ):
        self.use_vision = use_vision
        self.include_attributes = include_attributes
        self.llm: LLMEngine = llm
        self.conv: Conversation = Conversation()
        self.perception: BasePerception = perception

    @staticmethod
    def example() -> CaptchaDetection:
        return CaptchaDetection(
            has_captcha=True,
            captcha_type="image-based",
            description="A grid of 9 images is shown with the instruction 'Select all images containing traffic lights'",
        )

    def detection_message(self, step: TrajectoryStep) -> str:
        return f"""
Current page screenshot:
{self.perception.perceive(step.obs)}

Current page state:
{step.action.model_dump_json(exclude_unset=True)}
"""

    async def detect(
        self,
        step: TrajectoryStep,
    ) -> CaptchaDetection:
        """Detect if there is a captcha present in the current page screenshot"""
        self.conv.reset()
        system_prompt = chevron.render(system_rules, {"example": self.example().model_dump_json()})
        self.conv.add_system_message(content=system_prompt)
        self.conv.add_user_message(content=self.detection_message(step))

        answer: CaptchaDetection = await self.llm.structured_completion(self.conv.messages(), CaptchaDetection)
        return answer



from abc import ABC, abstractmethod


class BasePrompt(ABC):
    @abstractmethod
    def system(self) -> str:
        pass

    @abstractmethod
    def output_format_rules(self) -> str:
        pass

    @abstractmethod
    def select_action_rules(self) -> str:
        pass



from typing import Any, Literal

from loguru import logger
from notte_core.actions import ActionUnion, BaseAction, ClickAction, CompletionAction
from notte_core.common.config import config
from notte_sdk.types import render_agent_status
from pydantic import BaseModel, Field, field_serializer


class RelevantInteraction(BaseModel):
    """Interaction ids that can be relevant to the next actions"""

    id: str
    reason: str


class AgentState(BaseModel):
    """Current state of the agent"""

    previous_goal_status: Literal["success", "failure", "unknown"]
    previous_goal_eval: str
    page_summary: str
    relevant_interactions: list[RelevantInteraction]
    memory: str
    next_goal: str


class StepAgentOutput(BaseModel):
    state: AgentState
    actions: list[ActionUnion] = Field(min_length=1)

    @field_serializer("actions")
    def serialize_actions(self, actions: list[ActionUnion], _info: Any) -> list[dict[str, Any]]:
        return [action.model_dump_agent() for action in actions]

    @property
    def output(self) -> CompletionAction | None:
        if isinstance(self.actions[-1], CompletionAction):
            return self.actions[-1]
        return None

    def get_actions(self) -> list[BaseAction]:
        actions: list[BaseAction] = []
        # compute valid list of actions
        for i, _action in enumerate(self.actions):
            is_last = i == len(self.actions) - 1
            actions.append(_action)
            if not is_last and i >= config.max_actions_per_step:
                logger.warning(f"Max actions reached: {config.max_actions_per_step}. Skipping remaining actions.")
                break
            if not is_last and actions[-1].name() == ClickAction.name() and actions[-1].id.startswith("L"):
                logger.warning(f"Removing all actions after link click: {actions[-1].id}")
                # all actions after a link `L` should be removed from the list
                break
        return actions

    def log_state(self, colors: bool = True) -> list[tuple[str, dict[str, str]]]:
        action_str = ""
        for action in self.actions:
            action_str += f"   ▶ {action.name()} with id {action.id}"

        return render_agent_status(
            self.state.previous_goal_status,
            summary=self.state.page_summary,
            goal_eval=self.state.previous_goal_eval,
            memory=self.state.memory,
            next_goal=self.state.next_goal,
            action_str=action_str,
            colors=colors,
        )



import datetime as dt
import json
from collections.abc import Sequence
from enum import StrEnum
from pathlib import Path

import chevron
from notte_core.actions import (
    BaseAction,
    ClickAction,
    CompletionAction,
    FillAction,
    GotoAction,
    ScrapeAction,
)
from notte_core.errors.processing import InvalidInternalCheckError

system_prompt_dir = Path(__file__).parent / "prompts"


class ActionRegistry:
    """Union of all possible actions"""

    @staticmethod
    def render() -> str:
        """Returns a markdown formatted description of all available actions."""
        descriptions: list[str] = []

        for action_name, action_cls in BaseAction.ACTION_REGISTRY.items():
            try:
                # Get schema and safely remove common fields
                skip_keys = action_cls.non_agent_fields().difference(set(["description"]))
                sub_skip_keys = ["title", "$ref"]
                schema = {
                    k: {sub_k: sub_v for sub_k, sub_v in v.items() if sub_k not in sub_skip_keys}
                    for k, v in action_cls.model_json_schema()["properties"].items()
                    if k not in skip_keys
                }
                # schema['id'] = schema['id']['default']
                __description: dict[str, str] = schema.pop("description", "No description available")  # type: ignore[type-arg]
                if "default" not in __description:
                    raise InvalidInternalCheckError(
                        check=f"description should have a default value for {action_cls.__name__}",
                        url="unknown url",
                        dev_advice="This should never happen.",
                    )
                _description: str = __description["default"]
                # Format as: ActionName: {param1: {type: str, description: ...}, ...}
                description = f"""
* "{action_name}" : {_description}. Format:
```json
{json.dumps(schema)}
```
"""
                descriptions.append(description)
            except Exception as e:
                descriptions.append(f"Error getting schema for {action_cls.__name__}: {str(e)}")

        return "".join(descriptions)


class PromptType(StrEnum):
    SINGLE_ACTION = "single_action"
    MULTI_ACTION = "multi_action"

    def prompt_file(self) -> Path:
        match self:
            case PromptType.SINGLE_ACTION:
                return system_prompt_dir / "system_prompt_single_action.md"
            case PromptType.MULTI_ACTION:
                return system_prompt_dir / "system_prompt_multi_action.md"


class FalcoPrompt:
    def __init__(
        self,
        max_actions_per_step: int,
    ) -> None:
        multi_act = max_actions_per_step > 1
        prompt_type = PromptType.MULTI_ACTION if multi_act else PromptType.SINGLE_ACTION
        self.system_prompt: str = prompt_type.prompt_file().read_text()
        self.max_actions_per_step: int = max_actions_per_step

    @staticmethod
    def action_registry() -> str:
        return ActionRegistry.render()

    @staticmethod
    def _json_dump(actions: Sequence[BaseAction]) -> str:
        lines = ",\n  ".join([action.model_dump_agent_json() for action in actions])
        return "[\n  " + lines + "\n]"

    def example_form_filling(self) -> str:
        return self._json_dump(
            actions=[
                FillAction(id="I99", value="username"),
                FillAction(id="I101", value="password"),
                ClickAction(id="B1"),
            ]
        )

    def example_invalid_sequence(self) -> str:
        return self._json_dump(actions=[ClickAction(id="L1"), ClickAction(id="B4"), ClickAction(id="L2")])

    def example_navigation_and_extraction(self) -> str:
        return self._json_dump(
            [GotoAction(url="https://www.google.com"), ScrapeAction(instructions="Extract the search results")]
        )

    def completion_example(self) -> str:
        return self._json_dump([CompletionAction(success=True, answer="<answer to the task>")])

    def example_step(self) -> str:
        goal_eval = (
            "Analyze the current elements and the image to check if the previous goals/actions"
            " are successful like intended by the task. Ignore the action result. The website is the ground truth. "
            "Also mention if something unexpected happened like new suggestions in an input field. "
            "Shortly state why/why not"
        )
        return chevron.render(
            """
{
  "state": {
    "page_summary": "On the page are company a,b,c wtih their revenue 1,2,3.",
    "relevant_interactions": [{"id": "B2", "reason": "The button with id B2 represents search and I'm looking to search"}],
    "previous_goal_status": "success|failure|unknown",
    "previous_goal_eval": "{{goal_eval}}",
    "memory": "Description of what has been done and what you need to remember until the end of the task",
    "next_goal": "What needs to be done with the next actions"
  },
  "actions": [
   {
      "type: "one_action_type",
      // action-specific parameter
      ...
   }, // ... more actions in sequence ...
  ]
}
""",
            {"goal_eval": goal_eval},
        )

    def system(self) -> str:
        return chevron.render(
            self.system_prompt,
            {
                "timstamp": dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "max_actions_per_step": self.max_actions_per_step,
                "action_description": self.action_registry(),
                "example_form_filling": self.example_form_filling(),
                "example_step": self.example_step(),
                "completion_example": self.completion_example(),
                "completion_action_name": CompletionAction.name(),
                "goto_action_name": GotoAction.name(),
                "example_navigation_and_extraction": self.example_navigation_and_extraction(),
                "example_invalid_sequence": self.example_invalid_sequence(),
            },
        )

    def task(self, task: str):
        return f"""
Your ultimate task is: "{task}".
If you achieved your ultimate task, stop everything and use the done action in the next step to complete the task.
If not, continue as usual.
"""

    def new_task(self, task: str) -> str:
        return f"""
Your new ultimate task is: {task}.
Take the previous context into account and finish your new ultimate task.
"""

    def action_message(self) -> str:
        return """Given the previous information, start by reflecting on your last action. Then, summarize the current page and list relevant available interactions.
Absolutely do not under any circumstance list or pay attention to any id that is not explicitly found in the page.
From there, select the your next goal, and in turn, your next action.
    """



from abc import ABC, abstractmethod
from typing import Generic

from notte_core.actions import BaseAction, GotoAction
from notte_core.browser.observation import Observation, TrajectoryProgress
from notte_core.common.tracer import TStepAgentOutput
from pydantic import BaseModel, Field

from notte_agent.common.safe_executor import ExecutionStatus

ExecutionStepStatus = ExecutionStatus[BaseAction, Observation]


class TrajectoryStep(BaseModel, Generic[TStepAgentOutput]):
    agent_response: TStepAgentOutput
    results: list[ExecutionStepStatus]

    def observations(self) -> list[Observation]:
        return [result.output for result in self.results if result.output is not None]


def trim_message(message: str, max_length: int | None = None) -> str:
    if max_length is None or len(message) <= max_length:
        return message
    return f"...{message[-max_length:]}"


class TrajectoryHistory(BaseModel, ABC, Generic[TStepAgentOutput]):  # type: ignore[reportUnsafeMultipleInheritance]
    max_steps: int
    steps: list[TrajectoryStep[TStepAgentOutput]] = Field(default_factory=list)
    max_error_length: int | None = None

    @property
    def progress(self) -> TrajectoryProgress:
        return TrajectoryProgress(
            max_steps=self.max_steps,
            current_step=len(self.steps),
        )

    def reset(self) -> None:
        self.steps = []

    def perceive(self) -> str:
        steps = "\n".join([self.perceive_step(step, step_idx=i) for i, step in enumerate(self.steps)])
        return f"""
[Start of action execution history memory]
{steps or self.start_rules()}
[End of action execution history memory]
    """

    def start_rules(self) -> str:
        return f"""
No action executed so far...
Your first action should always be a `{GotoAction.name()}` action with a url related to the task.
You should reflect what url best fits the task you are trying to solve to start the task, e.g.
- flight search task => https://www.google.com/travel/flights
- go to reddit => https://www.reddit.com
- ...
ONLY if you have ABSOLUTELY no idea what to do, you can use `https://www.google.com` as the default url.
THIS SHOULD BE THE LAST RESORT.
"""

    def perceive_step_result(
        self,
        result: ExecutionStepStatus,
        include_ids: bool = False,
        include_data: bool = False,
    ) -> str:
        return self.perceive_execution_result(
            result, include_ids=include_ids, include_data=include_data, max_error_length=self.max_error_length
        )

    @staticmethod
    def perceive_execution_result(
        result: ExecutionStepStatus,
        include_ids: bool = False,
        include_data: bool = False,
        max_error_length: int | None = None,
    ) -> str:
        action = result.input
        id_str = f" with id={action.id}" if include_ids else ""
        if not result.success:
            err_msg = trim_message(result.message, max_error_length)
            return f"❌ action '{action.name()}'{id_str} failed with error: {err_msg}"
        success_msg = f"✅ action '{action.name()}'{id_str} succeeded: '{action.execution_message()}'"
        data = result.get().data
        if include_data and data is not None and data.structured is not None and data.structured.data is not None:
            return f"{success_msg}\n\nExtracted JSON data:\n{data.structured.data.model_dump_json()}"
        return success_msg

    @abstractmethod
    def perceive_step(
        self,
        step: TrajectoryStep[TStepAgentOutput],
        step_idx: int = 0,
        include_ids: bool = False,
        include_data: bool = True,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def add_output(self, output: TStepAgentOutput) -> None:
        raise NotImplementedError

    def add_step(self, step: ExecutionStepStatus) -> None:
        if len(self.steps) == 0:
            raise ValueError("Cannot add step to empty trajectory. Use `add_output` first.")
        else:
            if step.output is not None:
                step.output.progress = self.progress
            self.steps[-1].results.append(step)

    def observations(self) -> list[Observation]:
        obs = [obs for step in self.steps for obs in step.observations()]
        for o in obs:
            assert o.progress is not None, "Internal check: progress should be set for all observations"
        return obs

    def last_obs(self) -> Observation | None:
        for step in self.steps[::-1]:
            for step_result in step.results[::-1]:
                if step_result.success and step_result.output is not None:
                    return step_result.output
        return None



from typing import ClassVar, Literal

from notte_core.actions import ActionParameterValue, CompletionAction, GotoAction, ScrapeAction, StepAction
from typing_extensions import override

from notte_agent.common.parser import BaseParser, NotteStepAgentOutput


class GufoParser(BaseParser):
    observe_tag: ClassVar[str] = "observe"
    step_tag: ClassVar[str] = "execute-action"
    scrape_tag: ClassVar[str] = "scrape-data"
    done_tag: ClassVar[str] = "done"

    @override
    def example_format(self, endpoint: Literal["observe", "step", "scrape", "done", "error"]) -> str | None:
        match endpoint:
            case "observe":
                return f"""
<{self.observe_tag}>
{GotoAction(url="https://www.example.com").model_dump_agent_json()}
</{self.observe_tag}>
"""
            case "step":
                return f"""
<{self.step_tag}>
{
                    StepAction(
                        id="<YOUR_ACTION_ID>",
                        value=ActionParameterValue(name="<YOUR_PARAM_NAME>", value="<YOUR_PARAM_VALUE>"),
                    ).model_dump_agent_json()
                }
</{self.step_tag}>
"""
            case "scrape":
                return f"""
<{self.scrape_tag}>
{ScrapeAction(instructions="<YOUR_SCRAPING_INSTRUCTIONS | null to scrape the whole page>").model_dump_agent_json()}
</{self.scrape_tag}>
"""
            case "done":
                return f"""
<{self.done_tag}>
{CompletionAction(success=True, answer="<YOUR_ANSWER>").model_dump_agent_json()}
</{self.done_tag}>
"""
            case "error":
                return f"""
<{self.done_tag}>
{CompletionAction(success=False, answer="<REASON_FOR_FAILURE>").model_dump_agent_json()}
</{self.done_tag}>
"""

    @override
    def parse(self, text: str) -> NotteStepAgentOutput | None:
        url = self.search_pattern(text, GufoParser.observe_tag)
        action = self.search_pattern(text, GufoParser.step_tag)
        scrape = self.search_pattern(text, GufoParser.scrape_tag)
        output = self.search_pattern(text, GufoParser.done_tag)
        match (bool(url), bool(action), bool(scrape), bool(output)):
            case (True, False, False, False):
                return NotteStepAgentOutput(
                    observe=GotoAction.model_validate(self.parse_json(text, GufoParser.observe_tag))
                )
            case (False, True, False, False):
                return NotteStepAgentOutput(
                    step=StepAction.model_validate(self.parse_json(text, GufoParser.step_tag)),
                )
            case (False, False, True, False):
                return NotteStepAgentOutput(
                    scrape=ScrapeAction.model_validate(self.parse_json(text, GufoParser.scrape_tag))
                )
            case (False, False, False, True):
                return NotteStepAgentOutput(
                    completion=CompletionAction.model_validate(self.parse_json(text, GufoParser.done_tag))
                )
            case _:
                return None



from typing import final

from notte_core.browser.observation import Observation
from typing_extensions import override

from notte_agent.common.perception import BasePerception


@final
class GufoPerception(BasePerception):
    @override
    def perceive_metadata(self, obs: Observation) -> str:
        space_description = obs.space.description
        category: str = obs.space.category.value if obs.space.category is not None else ""
        return f"""
Webpage information:
- URL: {obs.metadata.url}
- Title: {obs.metadata.title}
- Description: {space_description or "No description available"}
- Timestamp: {obs.metadata.timestamp.strftime("%Y-%m-%d %H:%M:%S")}
- Page category: {category or "No category available"}
"""

    @override
    def perceive_data(
        self,
        obs: Observation,
    ) -> str:
        if not obs.has_data():
            raise ValueError("No scraping data found")
        return f"""
Here is some data that has been extracted from this page:
<data>
{obs.data.markdown if obs.data is not None else "No data available"}
</data>
"""

    @override
    def perceive_actions(self, obs: Observation) -> str:
        return f"""
Here are the available actions you can take on this page:
<actions>
{obs.space.markdown}
</actions>
"""

    @override
    def perceive(self, obs: Observation) -> str:
        return f"""
{self.perceive_metadata(obs).strip()}
{self.perceive_data(obs).strip() if obs.has_data() else ""}
{self.perceive_actions(obs).strip()}
"""



from typing import final

from notte_core.browser.observation import Observation
from typing_extensions import override

from notte_agent.common.perception import BasePerception


@final
class FalcoPerception(BasePerception):
    def __init__(
        self,
        include_step_info: bool = True,
        include_attributes: list[str] | None = None,
    ):
        self.include_attributes = include_attributes
        self.include_step_info = include_step_info

    @override
    def perceive_metadata(self, obs: Observation) -> str:
        if obs.progress is None:
            raise ValueError("Observation has no progress")
        return f"""
You will see the following only once. If you need to remember it and you dont know it yet, write it down in the memory.

* Current url: {obs.metadata.url}
* Current page title: {obs.metadata.title}
* Current date and time: {obs.metadata.timestamp.strftime("%Y-%m-%d %H:%M:%S")}
* Available tabs:
{obs.metadata.tabs}
* Current step: {obs.progress.current_step}/{obs.progress.max_steps}'
"""

    @override
    def perceive(self, obs: Observation) -> str:
        return f"""
[Relevant metadata]
{self.perceive_metadata(obs)}

[Interaction elements and context]
{self.perceive_actions(obs)}

[Data found in the page]
{self.perceive_data(obs)}
"""

    @override
    def perceive_actions(self, obs: Observation) -> str:
        px_above = obs.metadata.viewport.pixels_above
        px_below = obs.metadata.viewport.pixels_below

        more_above = f"... {px_above} pixels above - scroll or scrape content to see more ..."
        more_below = f"... {px_below} pixels below - scroll or scrape content to see more ..."

        space_description = obs.space.description

        return f"""
[Start of page]
{more_above if px_above > 0 else ""}
{space_description or "No content to display"}
{more_below if px_below > 0 else ""}
[End of page]

"""

    @override
    def perceive_data(self, obs: Observation, raw: bool = True) -> str:
        if not obs.has_data() or obs.data is None:
            return ""
        if raw:
            percieved_data = obs.data.markdown
        else:
            structured_data = obs.data.structured
            if structured_data is None or not structured_data.success or structured_data.data is None:
                error_msg = f" with error: {structured_data.error}" if structured_data is not None else ""
                return f"Scraping failed{error_msg}. Please try again with different instructions."
            percieved_data = structured_data.data.model_dump_json()

        return f"""
Data scraped from current page view:

{percieved_data or "No valid data to display"}
"""



from abc import ABC, abstractmethod

from notte_core.browser.observation import Observation


class BasePerception(ABC):
    @abstractmethod
    def perceive_metadata(self, obs: Observation) -> str:
        pass

    @abstractmethod
    def perceive_actions(self, obs: Observation) -> str:
        pass

    @abstractmethod
    def perceive_data(self, obs: Observation) -> str:
        pass

    @abstractmethod
    def perceive(self, obs: Observation) -> str:
        pass



from __future__ import annotations

from litellm import AllMessageValues
from notte_browser.session import TrajectoryStep
from notte_core.common.tracer import LlmUsageDictTracer
from notte_core.utils.webp_replay import ScreenshotReplay, WebpReplay
from pydantic import BaseModel
from typing_extensions import override

from notte_agent.common.trajectory_history import TrajectoryStep as AgentTrajectoryStep


class AgentResponse(BaseModel):
    success: bool
    answer: str
    session_trajectory: list[TrajectoryStep]
    agent_trajectory: list[AgentTrajectoryStep[BaseModel]]
    messages: list[AllMessageValues] | None = None
    llm_usage: list[LlmUsageDictTracer.LlmUsage]
    duration_in_s: float = -1

    @override
    def __str__(self) -> str:
        return (
            f"AgentResponse(success={self.success}, duration_in_s={round(self.duration_in_s, 2)}, answer={self.answer})"
        )

    def replay(self) -> WebpReplay:
        screenshots: list[bytes] = [
            obs.screenshot
            for step in self.agent_trajectory
            for obs in step.observations()
            if obs.screenshot is not None
        ]
        if len(screenshots) == 0:
            raise ValueError("No screenshots found in agent trajectory")
        return ScreenshotReplay.from_bytes(screenshots).get()

    @override
    def __repr__(self) -> str:
        return self.__str__()



from pathlib import Path

import chevron

from notte_agent.gufo.parser import GufoParser

system_prompt_file = Path(__file__).parent / "system.md"


class GufoPrompt:
    def __init__(self, parser: GufoParser):
        self.parser: GufoParser = parser
        self.system_prompt: str = system_prompt_file.read_text()

    def system(self, task: str, url: str | None = None) -> str:
        return chevron.render(self.system_prompt, {"task": task, "url": url or "the web"}, warn=True)

    def env_rules(self) -> str:
        return f"""
Hi there! I am the Notte web environment, and will help you navigate the internet.
# How it works:
* Provide me with a URL. I will respond with the actions you can take on that page.
* You are NOT allowed to provide me with more than one URL.
* Important: Make sure to use the **exact format** below when sending me a URL:
{self.parser.example_format("observe")}
> So, where would you like to go?
"""

    def completion_rules(self) -> str:
        return f"""
# How to format your answer when you're done
## Success answer
* If you're done, include you final answer in <{self.parser.done_tag}> tags.
* Don't forget to justify why your answer is correct and solves the task.
* Don't assume anything, just provide factual information backuped by the page you're on.
Format your answer as follows:
{self.parser.example_format("done")}

## Error answer
* If you feel stuck, remember that you are also allowed to use `Special Browser Actions` at any time to:
    * Go to a different url
    * Go back to the previous page
    * Refresh the current page
    * Scrape data from the page
    * Etc
* If you want to stop or you're unable to pursue your goal, format your answer as follows:
{self.parser.example_format("error")}
"""

    def select_action_rules(self) -> str:
        return f"""
# Next Action Selection
* Provide me with the ID of the action you want to take next.
* You are allowed to take only exactly ONE action from the list.
* You are ONLY allowed to pick actions from the latest list of actions!
* You are NOT allowed to pick actions from list of actions in previous messages!
* If the action is parameterized, provide the value for each parameter.
Use the exact following format:

{self.parser.example_format("step")}
"""



import typing
from collections.abc import Callable
from typing import Unpack

from loguru import logger
from notte_browser.session import NotteSession
from notte_browser.vault import VaultSecretsScreenshotMask
from notte_browser.window import BrowserWindow
from notte_core.actions import CompletionAction
from notte_core.common.config import NotteConfig
from notte_core.common.tracer import LlmUsageDictTracer
from notte_core.credentials.base import BaseVault
from notte_core.llms.engine import LLMEngine
from notte_sdk.types import AgentCreateRequestDict, AgentRunRequestDict
from pydantic import field_validator
from typing_extensions import override

from notte_agent.common.base import BaseAgent
from notte_agent.common.conversation import Conversation
from notte_agent.common.parser import NotteStepAgentOutput
from notte_agent.common.types import AgentResponse
from notte_agent.falco.agent import FalcoAgent
from notte_agent.gufo.parser import GufoParser
from notte_agent.gufo.perception import GufoPerception
from notte_agent.gufo.prompt import GufoPrompt


class GufoConfig(NotteConfig):
    enable_perception: bool = True
    auto_scrape: bool = True

    @field_validator("enable_perception")
    def check_perception(cls, value: bool) -> bool:
        if not value:
            raise ValueError("Perception should be enabled for gufo. Don't set this argument to `False`.")
        return value

    @field_validator("auto_scrape")
    def check_auto_scrape(cls, value: bool) -> bool:
        if not value:
            raise ValueError("Auto scrape should be enabled for gufo. Don't set this argument to `False`.")
        return value


class GufoAgent(BaseAgent):
    """
    A base agent implementation that coordinates between an LLM and the Notte environment.

    This class demonstrates how to build an agent that can:
    1. Maintain a conversation with an LLM
    2. Execute actions in the Notte environment
    3. Parse and format responses between the LLM and Notte

    To customize this agent:
    1. Implement your own Parser class to format observations and actions
    2. Modify the conversation flow in the run() method
    3. Adjust the think() method to handle LLM interactions
    4. Customize the ask_notte() method for your specific needs

    Args:
        task (str): The task description for the agent
        model (str): The LLM model identifier
        max_steps (int): Maximum number of steps before terminating
        headless (bool): Whether to run browser in headless mode
        parser (Parser | None): Custom parser for formatting interactions
    """

    def __init__(
        self,
        window: BrowserWindow,
        vault: BaseVault | None = None,
        step_callback: Callable[[str, NotteStepAgentOutput], None] | None = None,
        **data: Unpack[AgentCreateRequestDict],
    ) -> None:
        self.config: GufoConfig = GufoConfig.from_toml(**data)
        session = NotteSession(window=window, enable_perception=True)
        super().__init__(session=session)

        self.step_callback: Callable[[str, NotteStepAgentOutput], None] | None = step_callback
        self.tracer: LlmUsageDictTracer = LlmUsageDictTracer()
        self.vault: BaseVault | None = vault
        self.llm: LLMEngine = LLMEngine(model=self.config.reasoning_model, tracer=self.tracer)
        # Users should implement their own parser to customize how observations
        # and actions are formatted for their specific LLM and use case
        self.parser: GufoParser = GufoParser()
        self.prompt: GufoPrompt = GufoPrompt(self.parser)
        self.perception: GufoPerception = GufoPerception()
        self.conv: Conversation = Conversation()

        if self.vault is not None:
            # hide vault leaked credentials within llm inputs
            self.llm.structured_completion = self.vault.patch_structured_completion(0, self.vault.get_replacement_map)(  # pyright: ignore [reportAttributeAccessIssue]
                self.llm.structured_completion
            )

    async def reset(self):
        await self.session.areset()
        self.conv.reset()

    def output(self, answer: str, success: bool) -> AgentResponse:
        return AgentResponse(
            answer=answer,
            success=success,
            session_trajectory=self.session.trajectory,
            agent_trajectory=[],
            llm_usage=self.tracer.usage,
        )

    async def step(self, task: str) -> CompletionAction | None:
        # Processes the conversation history through the LLM to decide the next action.
        # logger.info(f"🤖 LLM prompt:\n{self.conv.messages()}")
        response: str = await self.llm.single_completion(self.conv.messages())
        self.conv.add_assistant_message(content=response)
        logger.info(f"🤖 LLM response:\n{response}")
        # Ask Notte to perform the selected action
        parsed_response = self.parser.parse(response)

        if parsed_response is None or parsed_response.action is None:
            self.conv.add_user_message(content=self.prompt.env_rules())
            return None

        if self.step_callback is not None:
            self.step_callback(task, parsed_response)

        if parsed_response.completion is not None:
            return parsed_response.completion
        action = parsed_response.action
        # Replace credentials if needed using the vault
        if self.vault is not None and self.vault.contains_credentials(action):
            locator = await self.session.locate(action)
            if locator is not None:
                action = self.vault.replace_credentials(
                    action,
                    await FalcoAgent.compute_locator_attributes(locator),
                    self.session.snapshot,
                )
        # Execute the action
        _ = await self.session.astep(action)
        obs = await self.session.aobserve()
        text_obs = self.perception.perceive(obs)
        self.conv.add_user_message(
            content=f"""
{text_obs}
{self.prompt.select_action_rules()}
{self.prompt.completion_rules()}
""",
            image=obs.screenshot if self.config.use_vision else None,
        )
        logger.info(f"🌌 Action successfully executed:\n{text_obs}")
        return None

    @override
    async def run(self, **kwargs: typing.Unpack[AgentRunRequestDict]) -> AgentResponse:
        """
        Main execution loop that coordinates between the LLM and Notte environment.

        This method shows a basic conversation flow. Consider customizing:
        1. The initial system prompt
        2. How observations are added to the conversation
        3. When and how to determine task completion
        4. Error handling and recovery strategies
        """
        task = kwargs.get("task")
        url = kwargs.get("url")
        logger.info(f"🚀 starting agent with task: {task} and url: {url}")
        system_msg = self.prompt.system(task, url)
        if self.vault is not None:
            system_msg += "\n" + self.vault.instructions()
        self.conv.add_system_message(content=system_msg)
        self.conv.add_user_message(self.prompt.env_rules())

        if self.vault is not None:
            self.session.window.screenshot_mask = VaultSecretsScreenshotMask(vault=self.vault)
        for i in range(self.config.max_steps):
            logger.info(f"> step {i}: looping in")
            output = await self.step(task=task)
            if output is not None:
                status = "😎 task completed sucessfully" if output.success else "👿 task failed"
                logger.info(f"{status} with answer: {output.answer}")
                return self.output(output.answer, output.success)
        # If the task is not done, raise an error
        error_msg = f"Failed to solve task in {self.config.max_steps} steps"
        logger.info(f"🚨 {error_msg}")
        return self.output(error_msg, False)
